import os
import numpy as np
import pandas as pd
from sklearn.preprocessing import MinMaxScaler
import tensorflow as tf
import tensorflow.python.keras.backend as K
from tensorflow.python.keras.optimizers import adam_v2
from tensorflow.python.keras.models import Model, model_from_json
from joblib import dump, load
from scipy.signal import find_peaks
from tensorflow.python.keras.layers import Input
from spectrum import Periodogram
from typing import List, Optional
from mtsa.models.ransyncoders_components.rancoders import RANCoders
from mtsa.models.ransyncoders_components.frequencycoder import FrequencyCoder
from mtsa.models.ransyncoders_components.sinusoidalcoder import SinusoidalCoder
from mtsa.models.ransyncoders_components.exceptions.parametererror import ParameterError

class RANSynCodersBase():
    def __init__(
            self,
            n_estimators: int,
            max_features: int,
            encoding_depth: int,
            latent_dim: int,
            decoding_depth: int,
            activation: str,
            output_activation: str,
            delta: float,
            synchronize: bool,
            force_synchronization: bool,
            min_periods: int,
            freq_init: Optional[List[float]],
            max_freqs: int,
            min_dist: int,
            trainable_freq: bool,
            bias: bool,
            sampling_rate: int,
            mono: bool,
            is_acoustic_data: bool,
            normal_classifier: int,
            abnormal_classifier: int,
        ):
        # Rancoders inputs:
        self.n_estimators = n_estimators
        self.max_features = max_features
        self.encoding_depth = encoding_depth
        self.latent_dim = latent_dim
        self.decoding_depth = decoding_depth
        self.activation = activation
        self.output_activation = output_activation
        self.delta = delta
        
        # Syncrhonization inputs
        self.synchronize = synchronize
        self.force_synchronization = force_synchronization
        self.min_periods = min_periods
        self.freq_init = freq_init  # in radians (angular frequency)
        self.max_freqs = max_freqs
        self.min_dist = min_dist
        self.trainable_freq = trainable_freq
        self.bias = bias
        self.sampling_rate = sampling_rate
        self.mono = mono
        self.is_acoustic_data = is_acoustic_data
        self.normal_classifier = normal_classifier
        self.abnormal_classifier = abnormal_classifier
        self.min_max_scaler = MinMaxScaler()
        # set all variables to default to float32
        tf.keras.backend.set_floatx('float32')
    
    #Region: Public methods      
    def fit(self, 
            X: np.ndarray,
            y, 
            timestamps_matrix: np.ndarray,
            learning_rate: float,
            batch_size: int, #old default value = 360
            epochs: int, #old default value = 100 
            freq_warmup: int,  # number of warmup epochs to prefit the frequency = 10 (old default value)
            sin_warmup: int,  # number of warmup epochs to prefit the sinusoidal representation = 10 (old default value)
            pos_amp: bool,  # whether to constraint amplitudes to be +ve only
            shuffle: bool
        ):
        self.timestamps_matrix = timestamps_matrix
        self.batch_size = batch_size
        self.experiment_dataframe = self.__get_experiment_dataframe()

        if self.is_acoustic_data:
            self.__train_model_with_audio_data(X, learning_rate, epochs, freq_warmup, sin_warmup, pos_amp) 
        else:
            self.__train_model_with_default_data(X, timestamps_matrix, learning_rate, epochs, freq_warmup, sin_warmup, pos_amp, shuffle)
             
    def predict(self, X: np.ndarray, time_matrix: np.ndarray, desync: bool = False):
        if self.is_acoustic_data:
            return self.__predict_with_audio_data(X, desync)
        return self.__predict_with_default_data(X, time_matrix, desync)
                       
    def build(self, input_shape, learning_rate: float, initial_stage: bool = False):
        x_in = Input(shape=(input_shape[-1],))  # created for either raw signal or synchronized signal
        optmizer = adam_v2.Adam(learning_rate = learning_rate)
        if initial_stage:
            freq_out = FrequencyCoder()(x_in)
            self.freqcoder = Model(inputs=x_in, outputs=freq_out)
            self.freqcoder.compile(optimizer=optmizer, loss=lambda y,f: quantile_loss(0.5, y,f))
        else:
            bounds_out = RANCoders(
                    n_estimators = self.n_estimators,
                    max_features = self.max_features,
                    encoding_depth = self.encoding_depth,
                    latent_dim = self.latent_dim,
                    decoding_depth = self.decoding_depth,
                    delta = self.delta,
                    activation = self.activation,
                    output_activation = self.output_activation,
                    name='rancoders'
                )(x_in)
            self.rancoders = Model(inputs=x_in, outputs=bounds_out)
            self.rancoders.compile(
                    optimizer=optmizer, 
                    loss=[lambda y,f: quantile_loss(1-self.delta, y,f), lambda y,f: quantile_loss(self.delta, y,f)]
            )  
            K.set_value(self.rancoders.optimizer.learning_rate, 0.01)
            if self.synchronize:
                t_in = Input(shape=(input_shape[-1],))
                sin_out = SinusoidalCoder(freq_init=self.freq_init, trainable_freq=self.trainable_freq)(t_in)
                self.sincoder = Model(inputs=t_in, outputs=sin_out)
                self.sincoder.compile(optimizer=optmizer, loss=lambda y,f: quantile_loss(0.5, y,f))
                
    def predict_desynchronize(self, x: np.ndarray, x_sync: np.ndarray, o_hi: np.ndarray, o_lo: np.ndarray):
        if self.synchronize:
            E = (o_hi + o_lo)/ 2  # expected values
            deviation = tf.expand_dims(x_sync, axis=1) - E  # input (synchronzied) deviation from expected
            deviation = self.desynchronize(deviation)  # desynchronize
            E = tf.expand_dims(x, axis=1) - deviation  # expected values in desynchronized form
            offset = (o_hi - o_lo) / 2  # this is the offet from the expected value
            offset = abs(self.desynchronize(offset))  # desynch
            o_hi, o_lo = E + offset, E - offset  # add bound displacement to expected values
            return o_hi.numpy(), o_lo.numpy()  
        else:
            raise ParameterError('synchronize', 'parameter not set correctly for this method')
    
    def desynchronize(self, e: np.ndarray):
        if self.synchronize:
            b = self.sincoder.layers[1].wb / self.sincoder.layers[1].freq  # phase shift(s)
            return e * tf.sin(
                self.sincoder.layers[1].freq[0] * ((np.pi / (2 * self.sincoder.layers[1].freq[0])) + b[:,0])
            ).numpy()
        else:
            raise ParameterError('synchronize', 'parameter not set correctly for this method')
    
    def score_samples(self, X):
        sins, synched, upper, lower = self.predict(X, self.timestamps_matrix)
        synched_tiles = np.tile(synched.reshape(synched.shape[0], 1, synched.shape[1]), (1, self.n_estimators, 1))
        result = np.where((synched_tiles < lower) | (synched_tiles > upper), self.abnormal_classifier, self.normal_classifier)
        
        if self.is_acoustic_data:
           return np.mean(result)
        return np.mean(np.mean(result, axis=1), axis=1) 
    
    def set_timestamps_matrix_to_predict(self, timestamps_matrix: np.ndarray):
        self.timestamps_matrix = timestamps_matrix
    
    def save(self, filepath: str = os.path.join(os.getcwd(), 'ransyncoders.z')):
        file = {'params': self.get_config()}
        if self.synchronize:
            file['frequencycoder'] = {'model': self.freqcoder.to_json(), 'weights': self.freqcoder.get_weights()}
            file['sinusoidalcoder'] = {'model': self.sincoder.to_json(), 'weights': self.sincoder.get_weights()}
        file['rancoders'] = {'model': self.rancoders.to_json(), 'weights': self.rancoders.get_weights()}
        dump(file, filepath, compress=True)
    
    @classmethod
    def load(cls, filepath: str = os.path.join(os.getcwd(), 'ransyncoders.z')):
        file = load(filepath)
        cls = cls()
        for param, val in file['params'].items():
            setattr(cls, param, val)
        if cls.synchronize:
            cls.freqcoder = model_from_json(file['frequencycoder']['model'], custom_objects={'FrequencyCoder': FrequencyCoder})
            cls.freqcoder.set_weights(file['frequencycoder']['weights'])
            cls.sincoder = model_from_json(file['sinusoidalcoder']['model'], custom_objects={'SinusoidalCoder': SinusoidalCoder})
            cls.sincoder.set_weights(file['sinusoidalcoder']['weights'])
        cls.rancoders = model_from_json(file['rancoders']['model'], custom_objects={'RANCoders': RANCoders})  
        cls.rancoders.set_weights(file['rancoders']['weights'])
        return cls
    
    def get_config(self):
        return self.experiment_dataframe
    #End region

    #Region: Private Methods
    def __train_model_with_default_data(self, X: np.ndarray, timestamps_matrix: np.ndarray, learning_rate, epochs, freq_warmup, sin_warmup, pos_amp, shuffle):
        
        if timestamps_matrix is None:
            timestamps_matrix = self.__get_time_matrix(X=X)
        
        with tf.device('/cpu:0'):
            dataset = tf.data.Dataset.from_tensor_slices((X.astype(np.float32), timestamps_matrix.astype(np.float32)))
            if shuffle:
                dataset = dataset.shuffle(buffer_size=X.shape[0]).batch(self.batch_size)
            else:
                dataset = dataset.batch(self.batch_size)
                
        self.__build_models(X, learning_rate)
                
        self.__train_ransyncoders_model(X, learning_rate, epochs, freq_warmup, sin_warmup, pos_amp, dataset)

    def __train_model_with_audio_data(self, X, learning_rate, epochs, freq_warmup, sin_warmup, pos_amp):
        
        datasets, X = self.__load_dataset(X)
        
        # build and compile models (stage 1)
        self.__build_models(X[0], learning_rate)
        
        for index, dataset in enumerate(datasets):
            self.__train_ransyncoders_model(X[index], learning_rate, epochs, freq_warmup, sin_warmup, pos_amp, dataset)

    def __train_ransyncoders_model(self, X, learning_rate, epochs, freq_warmup, sin_warmup, pos_amp, dataset):
        # pretraining step 1:
        self.__train_autoencoder(X, learning_rate, freq_warmup, dataset)
            
        # pretraining step 2:    
        self.__train_synchronization(sin_warmup, pos_amp, dataset)
                    
        # train anomaly detector   
        self.__train_anomaly_detector(learning_rate, epochs, dataset)
    
    def __train_anomaly_detector(self, learning_rate, epochs, dataset):
        with tf.device('/gpu:1'):
            for epoch in range(epochs):
                print("\nStart of epoch %d" % (epoch,))
                if self.synchronize:
                    for step, (x_batch, t_batch) in enumerate(dataset):
                            # Train the sine wave encoder
                        with tf.GradientTape() as tape:
                                # forward pass
                            s = self.sincoder(t_batch)
                                
                                # compute loss
                            s_loss = self.sincoder.loss(x_batch, s)  # median loss
                            
                            # retrieve gradients and update weights
                        grads = tape.gradient(s_loss, self.sincoder.trainable_weights)
                        self.sincoder.optimizer.apply_gradients(zip(grads, self.sincoder.trainable_weights))

                            # synchronize batch
                        b = self.sincoder.layers[1].wb / self.sincoder.layers[1].freq  # phase shift(s)
                        b_sync = b - tf.expand_dims(b[:,0], axis=-1)
                        th_sync = tf.expand_dims(
                                tf.expand_dims(self.sincoder.layers[1].freq, axis=0), axis=0
                            ) * (tf.expand_dims(t_batch, axis=-1) + tf.expand_dims(b_sync, axis=0))  # synchronized angle
                        e = (
                                x_batch - s
                            ) * tf.sin(
                                self.sincoder.layers[1].freq[0] * ((np.pi / (2 * self.sincoder.layers[1].freq[0])) - b[:,0])
                            )  # noise
                        x_batch_sync = tf.reduce_sum(
                                tf.expand_dims(self.sincoder.layers[1].amp, axis=0) * tf.sin(th_sync), axis=-1
                            ) + self.sincoder.layers[1].disp + e
                            
                            # train the rancoders
                        with tf.GradientTape() as tape:
                                # forward pass
                            o_hi, o_lo = self.rancoders(x_batch_sync)
                                
                                # compute losses
                            o_hi_loss = self.rancoders.loss[0](
                                    tf.tile(tf.expand_dims(x_batch_sync, axis=0), (self.n_estimators, 1, 1)), o_hi
                                )
                            o_lo_loss = self.rancoders.loss[1](
                                    tf.tile(tf.expand_dims(x_batch_sync, axis=0), (self.n_estimators, 1, 1)), o_lo
                                )
                            o_loss = o_hi_loss + o_lo_loss

                            # retrieve gradients and update weights
                        grads = tape.gradient(o_loss, self.rancoders.trainable_weights)
                        self.rancoders.optimizer.apply_gradients(zip(grads, self.rancoders.trainable_weights))

                        upper_bound_loss = tf.reduce_mean(o_hi_loss).numpy()
                        lower_bound_loss = tf.reduce_mean(o_lo_loss).numpy()
                    print(
                            "sine_loss:", tf.reduce_mean(s_loss).numpy(), 
                            "upper_bound_loss:", upper_bound_loss, 
                            "lower_bound_loss:", lower_bound_loss, 
                            end='\r'
                        )
                else:
                    for step, (x_batch, t_batch) in enumerate(dataset):
                        # train the rancoders
                        with tf.GradientTape() as tape:
                                # forward pass
                            o_hi, o_lo = self.rancoders(x_batch)
                                
                                # compute losses
                            o_hi_loss = self.rancoders.loss[0](
                                    tf.tile(tf.expand_dims(x_batch, axis=0), (self.n_estimators, 1, 1)), o_hi
                                )
                            o_lo_loss = self.rancoders.loss[1](
                                    tf.tile(tf.expand_dims(x_batch, axis=0), (self.n_estimators, 1, 1)), o_lo
                                )
                            o_loss = o_hi_loss + o_lo_loss

                            # retrieve gradients and update weights
                        grads = tape.gradient(o_loss, self.rancoders.trainable_weights)
                        self.rancoders.optimizer.apply_gradients(zip(grads, self.rancoders.trainable_weights))

                        upper_bound_loss = tf.reduce_mean(o_hi_loss).numpy()
                        lower_bound_loss = tf.reduce_mean(o_lo_loss).numpy()
                    print(
                            "upper_bound_loss:", upper_bound_loss, 
                            "lower_bound_loss:", lower_bound_loss, 
                            end='\r'
                        )

                self.experiment_dataframe.loc[len(self.experiment_dataframe)] = {"epoch_size": epochs,
                                                                                 "sampling_rate": self.sampling_rate,
                                                                                 "learning_rate": learning_rate,
                                                                                 "batch_size": self.batch_size,
                                                                                 "n_estimators": self.n_estimators,
                                                                                 "max_features": self.max_features,
                                                                                 "encoding_depth": self.encoding_depth,
                                                                                 "latent_dim": self.latent_dim,
                                                                                 "decoding_depth": self.decoding_depth,
                                                                                 "activation": self.decoding_depth,
                                                                                 "output_activation": self.decoding_depth,
                                                                                 "delta": self.delta,
                                                                                 "synchronize": self.synchronize,
                                                                                 "upper_bound_loss": upper_bound_loss,
                                                                                 "lower_bound_loss": lower_bound_loss,
                                                                                 "AUC_ROC": None,
                                                                                 "Confidence_interval_AUC_ROC": None
                                                                                }
    
    def __build_models(self, X, learning_rate):
        with tf.device('/cpu:0'):
            if self.synchronize:
                self.build(X.shape, initial_stage = True, learning_rate = learning_rate)
                if self.freq_init:
                    self.build(X.shape, learning_rate = learning_rate)
            else:
                self.build(X.shape, learning_rate = learning_rate)

    def __train_synchronization(self, sin_warmup, pos_amp, dataset):
        if sin_warmup > 0 and self.synchronize:
            with tf.device('/gpu:1'):
                for epoch in range(sin_warmup):
                    print("\nStart of sine representation pre-train epoch %d" % (epoch,))
                    for step, (x_batch, t_batch) in enumerate(dataset):
                            # Train the sine wave encoder
                        with tf.GradientTape() as tape:
                                # forward pass
                            s = self.sincoder(t_batch)
                                
                                # compute loss
                            s_loss = self.sincoder.loss(x_batch, s)  # median loss
                            
                            # retrieve gradients and update weights
                        grads = tape.gradient(s_loss, self.sincoder.trainable_weights)
                        self.sincoder.optimizer.apply_gradients(zip(grads, self.sincoder.trainable_weights))
                    print("sine_loss:", tf.reduce_mean(s_loss).numpy(), end='\r')
                
                # invert params (all amplitudes should either be -ve or +ve). Here we make them +ve
            if pos_amp:
                a_adj = tf.where(
                        self.sincoder.layers[1].amp[:,0] < 0, 
                        self.sincoder.layers[1].amp[:,0] * -1, 
                        self.sincoder.layers[1].amp[:,0]
                    )  # invert all -ve amplitudes
                wb_adj = tf.where(
                        self.sincoder.layers[1].amp[:,0] < 0, 
                        self.sincoder.layers[1].wb[:,0] + np.pi, 
                        self.sincoder.layers[1].wb[:,0]
                    )  # shift inverted waves by half cycle
                wb_adj = tf.where(
                        wb_adj > 2*np.pi, self.sincoder.layers[1].wb[:,0] - np.pi, wb_adj
                    )  # any cycle > freq must be reduced by half the cycle
                g_adj = tf.where(
                        self.sincoder.layers[1].amp[:,0] < 0, 
                        self.sincoder.layers[1].disp - a_adj, 
                        self.sincoder.layers[1].disp
                    )  # adjust the vertical displacements after reversing amplitude signs
                K.set_value(self.sincoder.layers[1].amp[:,0], a_adj)
                K.set_value(self.sincoder.layers[1].wb[:,0], wb_adj)
                K.set_value(self.sincoder.layers[1].disp, g_adj)

    def __train_autoencoder(self, X, learning_rate, freq_warmup, dataset):
        if freq_warmup > 0 and self.synchronize and not self.freq_init:
            with tf.device('/gpu:1'):
                for epoch in range(freq_warmup):
                    print("\nStart of frequency pre-train epoch %d" % (epoch,))
                    for step, (x_batch, t_batch) in enumerate(dataset):
                            # Prefit the oscillation encoder
                        with tf.GradientTape() as tape:
                                # forward pass
                            z, x_pred = self.freqcoder(x_batch)
                                
                                # compute loss
                            x_loss = self.freqcoder.loss(x_batch, x_pred)  # median loss
                            
                            # retrieve gradients and update weights
                        grads = tape.gradient(x_loss, self.freqcoder.trainable_weights)
                        self.freqcoder.optimizer.apply_gradients(zip(grads, self.freqcoder.trainable_weights))
                    print("pre-reconstruction_loss:", tf.reduce_mean(x_loss).numpy(), end='\r')
                    
                # estimate dominant frequency
            with tf.device('/cpu:0'):
                z = self.freqcoder(X)[0].numpy().reshape(-1)  # must be done on full unshuffled series
                z = ((z - z.min()) / (z.max() - z.min())) * (1 - -1) + -1  #  scale between -1 & 1
                p = Periodogram(z, sampling=1)
                p()
                peak_idxs = find_peaks(p.psd, distance=self.min_dist, height=(0, np.inf))[0]
                peak_order = p.psd[peak_idxs].argsort()[-self.min_periods-self.max_freqs:][::-1]  # max PSDs found
                peak_idxs = peak_idxs[peak_order]
            if peak_idxs[0] < self.min_periods and not self.force_synchronization:
                self.synchronize = False
                print('no common oscillations found, switching off synchronization attempts')
            elif max(peak_idxs[:self.min_periods]) >= self.min_periods:
                idxs = peak_idxs[peak_idxs >= self.min_periods]
                peak_freqs = [p.frequencies()[idx] for idx in idxs[:min(len(idxs), self.max_freqs)]]
                self.freq_init = [2 * np.pi * f for f in peak_freqs]
                print('found common oscillations at period(s) = {}'.format([1 / f for f in peak_freqs]))
            else:
                self.synchronize = False
                print('no common oscillations found, switching off synchronization attempts')
                
                # build and compile models (stage 2)
            with tf.device('/cpu:0'):
                self.build(X.shape, learning_rate = learning_rate)
    
    def __normalize_data(self, X: np.ndarray, training_step: bool):
        if(training_step):
            return self.min_max_scaler.fit_transform(X)
        return self.min_max_scaler.transform(X)
        
    def __get_time_matrix(self, X):
        #each line represents an instant of time in each time series
        total_time_in_seconds, time_serie_time_instant = self.__calculate_time_serie_time_instatnt(X)
        time_matrix = np.arange(start=0, stop=total_time_in_seconds, step=time_serie_time_instant)
        return np.tile(time_matrix.reshape(-1, 1), (1, X.shape[1]))

    def __calculate_time_serie_time_instatnt(self, X):
        X_length = X.shape[0]
        #REMAKE
        dividend = self.sampling_rate
        if X_length < self.sampling_rate:
            dividend = X_length
            
        total_time_in_seconds = X_length/dividend
        time_serie_time_instant = total_time_in_seconds/X_length
        return total_time_in_seconds,time_serie_time_instant

    def __load_dataset(self, X: np.ndarray, training_step: bool = True):
        datasets = []
        x_2D = []
        
        for x in X:
            x = self.__transform_X(x)
            time_matrix = self.__get_time_matrix(x)
            x_normalized = self.__normalize_data(x, training_step)
            x_2D.append(x_normalized)
            with tf.device('/cpu:0'):
                dataset = tf.data.Dataset.from_tensor_slices((x_normalized.astype(np.float32), time_matrix.astype(np.float32)))
                datasets.append(dataset.batch(self.batch_size))  
                    
        return datasets, x_2D

    def __transform_X(self, x):
        if self.mono and (not self.is_acoustic_data):
            x = x.reshape(-1, 1)
        else:
            x = x.T
        return x
    
    def __get_experiment_dataframe(self):
        parameters_columns = ["epoch_size",
                              "sampling_rate",
                              "learning_rate",
                              "batch_size",
                              "n_estimators",
                              "max_features",
                              "encoding_depth",
                              "latent_dim",
                              "decoding_depth",
                              "activation",
                              "output_activation",
                              "delta",
                              "synchronize",
                              "upper_bound_loss",
                              "lower_bound_loss",
                              "AUC_ROC",
                              "Confidence_interval_AUC_ROC"
                            ]               
        return pd.DataFrame(columns=parameters_columns)
                
    def __predict_with_audio_data(self, X: np.ndarray, desync):
        
        datasets, X = self.__load_dataset(X, False)
        batches = int(np.ceil(X[0].shape[0] / self.batch_size))
        
        return self.__model_predict(datasets[0], batches, desync)
            
    def __predict_with_default_data(self, X: np.ndarray, timestamps_matrix: np.ndarray, desync):
        with tf.device('/gpu:1'):
            if timestamps_matrix is None:
                timestamps_matrix = self.__get_time_matrix(X=X)
                
            dataset = tf.data.Dataset.from_tensor_slices((X.astype(np.float32), timestamps_matrix.astype(np.float32)))
            dataset = dataset.batch(self.batch_size)
            batches = int(np.ceil(X.shape[0] / self.batch_size))
            return self.__model_predict(dataset, batches, desync)
           
    def __model_predict(self, dataset, batches, desync):
        with tf.device('/gpu:1'):
            if self.synchronize:
                s, x_sync, o_hi, o_lo = [None] * batches, [None] * batches, [None] * batches, [None] * batches
                for step, (x_batch, t_batch) in enumerate(dataset):
                    s_i = self.sincoder(t_batch).numpy()
                    b = self.sincoder.layers[1].wb / self.sincoder.layers[1].freq  # phase shift(s)
                    b_sync = b - tf.expand_dims(b[:,0], axis=-1)
                    th_sync = tf.expand_dims(
                        tf.expand_dims(self.sincoder.layers[1].freq, axis=0), axis=0
                    ) * (tf.expand_dims(t_batch, axis=-1) + tf.expand_dims(b_sync, axis=0))  # synchronized angle
                    e = (
                        x_batch - s_i
                    ) * tf.sin(
                        self.sincoder.layers[1].freq[0] * ((np.pi / (2 * self.sincoder.layers[1].freq[0])) - b[:,0])
                    )  # noise
                    x_sync_i = (tf.reduce_sum(
                        tf.expand_dims(self.sincoder.layers[1].amp, axis=0) * tf.sin(th_sync), axis=-1
                    ) + self.sincoder.layers[1].disp + e).numpy()  
                    o_hi_i, o_lo_i = self.rancoders(x_sync_i)
                    o_hi_i, o_lo_i = tf.transpose(o_hi_i, [1,0,2]).numpy(), tf.transpose(o_lo_i, [1,0,2]).numpy()
                    if desync:
                        o_hi_i, o_lo_i = self.predict_desynchronize(x_batch, x_sync_i, o_hi_i, o_lo_i)
                    s[step], x_sync[step], o_hi[step], o_lo[step]  = s_i, x_sync_i, o_hi_i, o_lo_i
                return (
                    np.concatenate(s, axis=0), 
                    np.concatenate(x_sync, axis=0), 
                    np.concatenate(o_hi, axis=0), 
                    np.concatenate(o_lo, axis=0)
                )
            else:
                o_hi, o_lo = [None] * batches, [None] * batches
                for step, (x_batch, t_batch) in enumerate(dataset):
                    o_hi_i, o_lo_i = self.rancoders(x_batch)
                    o_hi_i, o_lo_i = tf.transpose(o_hi_i, [1,0,2]).numpy(), tf.transpose(o_lo_i, [1,0,2]).numpy()
                    o_hi[step], o_lo[step]  = o_hi_i, o_lo_i
                return np.concatenate(o_hi, axis=0), np.concatenate(o_lo, axis=0)
    #End region
    
# Loss function
def quantile_loss(q, y, f):
    e = (y - f)
    return K.mean(K.maximum(q*e, (q-1)*e), axis=-1)


from tensorflow.python.keras.layers import Dense, Layer

class FrequencyCoder(Layer):
    """ 
    Encode multivariate to a latent space of size 1 for extracting common oscillations in the series (similar to finding PCA).
    """
    def __init__(self, **kwargs):
        super(FrequencyCoder, self).__init__(**kwargs)
        self.kwargs = kwargs
        
    def build(self, input_shape):
        self.latent = Dense(1, activation='linear')
        self.decoder = Dense(input_shape[-1], activation='linear')
    
    def call(self, inputs):
        z = self.latent(inputs)
        x_pred = self.decoder(z)
        return z, x_pred
    
    def get_config(self):
        base_config = super(FrequencyCoder, self).get_config()
        return dict(list(base_config.items()))


import os
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

class LSTM_FC(nn.Module):
    def __init__(self,
                 input_dim=32,
                 num_classes=1, 
                 num_hidden_nodes=8
    ):

        super(LSTM_FC, self).__init__()
        self.input_dim = input_dim
        self.num_classes = num_classes
        self.num_hidden_nodes = num_hidden_nodes
        self.encoder = nn.LSTM(input_size=self.input_dim, hidden_size=self.num_hidden_nodes,
                                num_layers=1, batch_first=True)
        self.fc = nn.Linear(self.num_hidden_nodes, self.num_classes)
        activ = nn.ReLU(True)

    def forward(self, input):
        features = self.encoder(input)[0][:,-1,:]
        # pdb.set_trace()
        logits = self.fc(features)
        return logits

    def half_forward_start(self, x):
        features = self.encoder(x)[0][:,-1,:]
        return features

    def half_forward_end(self, x):
        logits = self.fc(x)
        return logits

class DROCCTrainer:
    """
    Trainer class that implements the DROCC algorithm proposed in
    https://arxiv.org/abs/2002.12718
    """

    def __init__(self, model, optimizer, lamda, radius, gamma, device):
        """Initialize the DROCC Trainer class
        Parameters
        ----------
        model: Torch neural network object
        optimizer: Total number of epochs for training.
        lamda: Weight given to the adversarial loss
        radius: Radius of hypersphere to sample points from.
        gamma: Parameter to vary projection.
        device: torch.device object for device to use.
        """     
        self.model = model
        self.optimizer = optimizer
        self.lamda = lamda
        self.radius = radius
        self.gamma = gamma
        self.device = device

    def train(self, train_loader, lr_scheduler, total_epochs, save_path, name,
                only_ce_epochs=5, ascent_step_size=0.001, ascent_num_steps=50):
        """Trains the model on the given training dataset with periodic 
        evaluation on the validation dataset.
        Parameters
        ----------
        train_loader: Dataloader object for the training dataset.
        val_loader: Dataloader object for the validation dataset.
        learning_rate: Initial learning rate for training.
        total_epochs: Total number of epochs for training.
        only_ce_epochs: Number of epochs for initial pretraining.
        ascent_step_size: Step size for gradient ascent for adversarial 
                          generation of negative points.
        ascent_num_steps: Number of gradient ascent steps for adversarial 
                          generation of negative points.
        metric: Metric used for evaluation (AUC / F1).
        """
        self.ascent_num_steps = ascent_num_steps
        self.ascent_step_size = ascent_step_size
        for epoch in range(total_epochs): 
            #Make the weights trainable
            self.model.train()
            
            #Placeholder for the respective 2 loss values
            epoch_adv_loss = 0.0  #AdvLoss
            epoch_ce_loss = 0.0  #Cross entropy Loss
            
            batch_idx = -1
            for data in train_loader:
                batch_idx += 1
                data = data.to(self.device)
                target = torch.ones([data.shape[0]], dtype=torch.float32).to(self.device)

                data = torch.transpose(data, dim0=1, dim1=2)
                data = data.reshape(data.shape[0], data.shape[1], data.shape[2]*data.shape[3])

                self.optimizer.zero_grad()
                
                # Extract the logits for cross entropy loss
                logits = self.model(data)
                logits = torch.squeeze(logits, dim = 1)
                ce_loss = F.binary_cross_entropy_with_logits(logits, target)
                # Add to the epoch variable for printing average CE Loss
                epoch_ce_loss += ce_loss.item()

                '''
                Adversarial Loss is calculated only for the positive data points (label==1).
                '''
                if  epoch >= only_ce_epochs:
                    data = data[target == 1]
                    # AdvLoss 
                    adv_loss = self.one_class_adv_loss(data)
                    epoch_adv_loss += adv_loss.item()

                    loss = ce_loss + adv_loss * self.lamda
                else: 
                    # If only CE based training has to be done
                    loss = ce_loss
                
                # Backprop
                loss.backward()
                self.optimizer.step()
            lr_scheduler.step()
                    
            epoch_ce_loss = epoch_ce_loss/(batch_idx + 1)  #Average CE Loss
            epoch_adv_loss = epoch_adv_loss/(batch_idx + 1) #Average AdvLoss

            print('Epoch: {}, CE Loss: {}, AdvLoss: {}'.format(
                epoch, epoch_ce_loss, epoch_adv_loss))
            self.save(os.path.join(save_path, "{}_{}.pt".format(name, epoch)))
    def test(self, test_loader):
        """Evaluate the model on the given test dataset.
        Parameters
        ----------
        test_loader: Dataloader object for the test dataset.
        metric: Metric used for evaluation (AUC / F1).
        """        
        self.model.eval()
        scores = []
        with torch.no_grad():
            for data in test_loader:
                data = data.to(self.device)

                data = torch.transpose(data, dim0=1, dim1=2)
                data = data.reshape(data.shape[0], data.shape[1], data.shape[2]*data.shape[3])


                logits = self.model(data).cpu().numpy()
                scores.append(logits)
        scores = -np.concatenate(scores)

        return scores
        
    
    def one_class_adv_loss(self, x_train_data):
        """Computes the adversarial loss:
        1) Sample points initially at random around the positive training
            data points
        2) Gradient ascent to find the most optimal point in set N_i(r) 
            classified as +ve (label=0). This is done by maximizing 
            the CE loss wrt label 0
        3) Project the points between spheres of radius R and gamma * R 
            (set N_i(r))
        4) Pass the calculated adversarial points through the model, 
            and calculate the CE loss wrt target class 0
        
        Parameters
        ----------
        x_train_data: Batch of data to compute loss on.
        """
        batch_size = len(x_train_data)
        # Randomly sample points around the training data
        # We will perform SGD on these to find the adversarial points
        x_adv = torch.randn(x_train_data.shape).to(self.device).detach().requires_grad_()
        x_adv_sampled = x_adv + x_train_data

        for step in range(self.ascent_num_steps):
            with torch.enable_grad():

                new_targets = torch.zeros(batch_size, 1).to(self.device)
                new_targets = torch.squeeze(new_targets)
                new_targets = new_targets.to(torch.float)
                
                logits = self.model(x_adv_sampled)         
                logits = torch.squeeze(logits, dim = 1)
                new_loss = F.binary_cross_entropy_with_logits(logits, new_targets)

                grad = torch.autograd.grad(new_loss, [x_adv_sampled])[0]
                grad_norm = torch.norm(grad, p=2, dim = tuple(range(1, grad.dim())))
                grad_norm = grad_norm.view(-1, *[1]*(grad.dim()-1))
                grad_normalized = grad/grad_norm 
            with torch.no_grad():
                x_adv_sampled.add_(self.ascent_step_size * grad_normalized)

            if (step + 1) % 10==0:
                # Project the normal points to the set N_i(r)
                h = x_adv_sampled - x_train_data
                norm_h = torch.sqrt(torch.sum(h**2, 
                                                dim=tuple(range(1, h.dim()))))
                alpha = torch.clamp(norm_h, self.radius, 
                                    self.gamma * self.radius).to(self.device)
                # Make use of broadcast to project h
                proj = (alpha/norm_h).view(-1, *[1] * (h.dim()-1))
                h = proj * h
                x_adv_sampled = x_train_data + h  #These adv_points are now on the surface of hyper-sphere

        adv_pred = self.model(x_adv_sampled)
        adv_pred = torch.squeeze(adv_pred, dim=1)
        adv_loss = F.binary_cross_entropy_with_logits(adv_pred, (new_targets * 0))

        return adv_loss

    def save(self, path):
        torch.save(self.model.state_dict(),path)

    def load(self, path):
        self.model.load_state_dict(torch.load(path))


from apache_beam.io.avroio import WriteToAvro

import apache_beam as beam
from apache_beam.pipeline import PipelineOptions

output = "/data/output_test/"

schema = {
    "namespace": "mtsa",
    "name": "ExperimentResult", 
    "type": "record",
    "fields": [
        {"name": "len", "type": "int"}
    ]
} 

def main(argv=None):
        
    
    with beam.Pipeline() as pipeline:

        runs = pipeline | beam.Create([
            "Hi",
            "This is a line",
            "To be written"
        ])

        params_individual =  runs | "individual" >> beam.Map(lambda x: {"len": len(x)})

        # write = params_individual | WriteToAvro(f"{output}test")
        write = params_individual | WriteToAvro(f"{output}test.avro", schema=schema)



if __name__ == "__main__":
    main()


from sklearn.base import BaseEstimator, TransformerMixin
import librosa as lib
from functools import reduce

NORMAL = 1
ABNORMAL = 0
class Wav2Array(BaseEstimator, TransformerMixin):
    def __init__(self, 
                 sampling_rate=None,
                 mono=True
                 ):
        self.sampling_rate = sampling_rate
        self.mono = mono

    def fit(self, root, y=None, **fit_params):
        return self

    def transform(self, X, y=None, **fit_params):
        def get_array(f):
            if self.sampling_rate:
                s, _ = lib.load(f, sr=self.sampling_rate, mono=self.mono)
                return s
            else:
                s, _ = lib.load(f, mono=self.mono)
                return s
        np.random.seed(0)
        Xt = np.array(list(map(get_array, X)))
        return Xt

class Demux2Array(BaseEstimator, TransformerMixin):
    def __init__(self, 
                 channel=0
                 ):
        self.channel = channel

    def fit(self, root, y=None, **fit_params):
        return self

    def transform(self, X, y=None, **fit_params):
        def get_array(Xi):
            # multi_channel_data, sr = file_load(wav_name)
            if Xi.ndim <= 1:
                return Xi
            return np.array(Xi)[self.channel, :]
        
        Xt = np.array(list(map(get_array, X)))
        return Xt

from sklearn.model_selection import BaseShuffleSplit
from sklearn.utils.validation import check_random_state 
from sklearn.model_selection._split import _validate_shuffle_split
import numpy as np
class AbnormalSplit(BaseShuffleSplit):

    def __init__(
        self, n_splits=10, random_state=None
    ):
        super().__init__(
            n_splits=n_splits,
            random_state=random_state
        )

    def _iter_indices(self, X, y=None, groups=None):
        NORMAL = 1
        ABNORMAL = 0
        ind_abnormal = np.where(y == ABNORMAL)[0]
        ind_normal = np.where(y == NORMAL)[0]

        rng = check_random_state(self.random_state)
        for i in range(self.n_splits):
            # random partition
            permutation_normal = rng.permutation(ind_normal)
            ind_test_normal = permutation_normal[0:len(ind_abnormal)]
            ind_train_normal = permutation_normal[len(ind_abnormal):]

            ind_train = ind_train_normal
            
            ind_test = np.concatenate([ind_test_normal, ind_abnormal])
            yield ind_train, ind_test

import os
import glob
# from functools import reduce
def get_files_from_path(path):
    pattern = "*.wav"
    path_wav = lambda signal_class: glob.glob(os.path.join(path, signal_class, pattern))
    return path_wav(path)

def get_X_y_from_normal_abnormal(normal, abnormal):
    y0 = NORMAL * np.ones(len(normal))
    y1 = ABNORMAL * np.ones(len(abnormal))
    X = np.concatenate([normal, abnormal])
    y = np.concatenate([y0, y1])
    return X, y
    
def get_files_from_path_classes(path):
    X0 = get_files_from_path(os.path.join(path, "normal"))
    X1 = get_files_from_path(os.path.join(path, "abnormal"))
    X, y = get_X_y_from_normal_abnormal(X0, X1)
    return X, y


def files_train_test_split(path, random_state=None):
    X, y = get_files_from_path_classes(path)
    ind_train, ind_test = next(AbnormalSplit(random_state=random_state,n_splits=1).split(X, y))
    X_train, X_test, y_train, y_test = X[ind_train], X[ind_test], y[ind_train], y[ind_test]
    return X_train, X_test, y_train, y_test

def files_train_test_split_dcase2020_task2(path, pattern=''):
    """
    class_id_sample.wav
    
    :Examples:
    normal_id_00_00000110.wav
     
    """
    
    X_train_normal = glob.glob(os.path.join(path, 'train', f'normal_{pattern}*'))
    X_train_abnormal = glob.glob(os.path.join(path, 'train', f'anomaly_{pattern}*'))
    
    X_test_abnormal = glob.glob(os.path.join(path, 'test', f'anomaly_{pattern}*'))
    X_test_normal = glob.glob(os.path.join(path, 'test', f'normal_{pattern}*'))
    
    X_train, y_train = get_X_y_from_normal_abnormal(X_train_normal, X_train_abnormal)
    X_test, y_test = get_X_y_from_normal_abnormal(X_test_normal, X_test_abnormal)
    
    return X_train, X_test, y_train, y_test


def files_train_test_split_combined(paths):
    def reduce_data(d1, d2):
        data = zip(d1, d2)
        data = list(data)
        X_train, X_test, y_train, y_test = [np.concatenate([d1, d2]) for (d1,d2) in data]
        return X_train, X_test, y_train, y_test
    data = list(map(files_train_test_split, paths))
    X_train, X_test, y_train, y_test = reduce(reduce_data, data)
    return X_train, X_test, y_train, y_test


import torch
import torch.nn as nn
from functools import partial

class RecurrentEncoder(nn.Module):
    """Recurrent encoder"""

    def __init__(self, n_features, latent_dim, rnn):
        super().__init__()

        self.rec_enc1 = rnn(n_features, latent_dim, batch_first=True)

    def forward(self, x):
        _, h_n = self.rec_enc1(x)

        return h_n

class RecurrentDecoder(nn.Module):
    """Recurrent decoder for RNN and GRU"""

    def __init__(self, latent_dim, n_features, rnn_cell, device):
        super().__init__()

        self.n_features = n_features
        self.device = device
        self.rec_dec1 = rnn_cell(n_features, latent_dim)
        self.dense_dec1 = nn.Linear(latent_dim, n_features)

    def forward(self, h_0, seq_len):
        # Initialize output
        x = torch.tensor([], device = self.device)

        # Squeezing
        h_i = h_0.squeeze()

        # Reconstruct first element with encoder output
        x_i = self.dense_dec1(h_i)

        # Reconstruct remaining elements
        for i in range(0, seq_len):
            h_i = self.rec_dec1(x_i, h_i)
            x_i = self.dense_dec1(h_i)
            x = torch.cat([x, x_i], axis=1)

        return x.view(-1, seq_len, self.n_features)


class RecurrentDecoderLSTM(nn.Module):
    """Recurrent decoder LSTM"""

    def __init__(self, latent_dim, n_features, rnn_cell, device):
        super().__init__()

        self.n_features = n_features
        self.device = device
        self.rec_dec1 = rnn_cell(n_features, latent_dim)
        self.dense_dec1 = nn.Linear(latent_dim, n_features)

    def forward(self, h_0, seq_len):
        # Initialize output
        x = torch.tensor([], device = self.device)

        # Squeezing
        h_i = [h.squeeze() for h in h_0]

        # Reconstruct first element with encoder output
        x_i = self.dense_dec1(h_i[0])

        # Reconstruct remaining elements
        for i in range(0, seq_len):
            h_i = self.rec_dec1(x_i, h_i)
            x_i = self.dense_dec1(h_i[0])
            x = torch.cat([x, x_i], axis = 1)

        return x.view(-1, seq_len, self.n_features)


class RecurrentAE(nn.Module):
    """Recurrent autoencoder"""

    def __init__(self, n_features, latent_dim, device):
        super().__init__()

        # Encoder and decoder argsuration
        self.rnn, self.rnn_cell = nn.LSTM, nn.LSTMCell
        self.decoder = RecurrentDecoderLSTM
        self.latent_dim = latent_dim
        self.n_features = n_features
        self.device = device

        # Encoder and decoder
        self.encoder = RecurrentEncoder(self.n_features, self.latent_dim, self.rnn)
        self.decoder = self.decoder(self.latent_dim, self.n_features, self.rnn_cell, self.device)

    def forward(self, x):
        # x: N X K X L X D 
        seq_len = x.shape[1]
        h_n = self.encoder(x)
        out = self.decoder(h_n, seq_len)

        return torch.flip(out, [1])

# %%



import numpy as np
import tensorflow as tf
from tensorflow.python.keras.layers import Layer
from mtsa.models.ransyncoders_components.decoder import Decoder
from mtsa.models.ransyncoders_components.encoder import Encoder

class RANCoders(Layer):
    def __init__(
            self, 
            n_estimators: int = 100,
            max_features: int = 3,
            encoding_depth: int = 2,
            latent_dim: int = 2, 
            decoding_depth: int = 2,
            delta: float = 0.05,
            activation: str = 'linear',
            output_activation: str = 'linear',
            **kwargs,
    ):
        super(RANCoders, self).__init__(**kwargs)
        self.n_estimators = n_estimators
        self.max_features = max_features
        self.encoding_depth = encoding_depth
        self.latent_dim = latent_dim
        self.decoding_depth = decoding_depth
        self.delta = delta
        self.activation = activation
        self.output_activation = output_activation
        self.kwargs = kwargs
        
    def build(self, input_shape):
        assert(input_shape[-1] >= self.max_features)
        self.encoders = {
            'encoder_{}'.format(i): Encoder(
                self.latent_dim, self.activation, depth=self.encoding_depth,
            ) for i in range(self.n_estimators)
        }
        self.decoders_upper = {
            'decoder_hi_{}'.format(i): Decoder(
                input_shape[-1], self.activation, self.output_activation, self.decoding_depth
            ) for i in range(self.n_estimators)
        }
        self.decoders_lower = {
            'decoder_lo_{}'.format(i): Decoder(
                input_shape[-1], self.activation, self.output_activation, self.decoding_depth
            ) for i in range(self.n_estimators)
        }
        self.randsamples = tf.Variable(
            np.concatenate(
                [
                    np.random.choice(
                        input_shape[-1], replace=False, size=(1, self.max_features),
                    ) for i in range(self.n_estimators)
                ]
            ), trainable=False
        )  # the feature selector (bootstrapping)
        
    def call(self, inputs):
        z = {
            'z_{}'.format(i): self.encoders['encoder_{}'.format(i)](
                tf.gather(inputs, self.randsamples[i], axis=-1)
            ) for i in range(self.n_estimators)
        }
        w_hi = {
            'w_{}'.format(i): self.decoders_upper['decoder_hi_{}'.format(i)](
                z['z_{}'.format(i)]
            ) for i in range(self.n_estimators)
        }
        w_lo = {
            'w_{}'.format(i): self.decoders_lower['decoder_lo_{}'.format(i)](
                z['z_{}'.format(i)]
            ) for i in range(self.n_estimators)
        }
        o_hi = tf.concat([tf.expand_dims(i, axis=0) for i in w_hi.values()], axis=0)  
        o_lo = tf.concat([tf.expand_dims(i, axis=0) for i in w_lo.values()], axis=0)
        return o_hi, o_lo
    
    def get_config(self):
        base_config = super(RANCoders, self).get_config()
        config = {
            "n_estimators": self.n_estimators,
            "max_features": self.max_features,
            "encoding_depth": self.encoding_depth,
            "latent_dim": self.latent_dim,
            "decoding_depth": self.decoding_depth,
            "delta": self.delta,
            "activation": self.activation,
            "output_activation": self.output_activation,
        }
        return dict(list(base_config.items()) + list(config.items()))


from sklearn import metrics
import torch.nn as nn
import numpy as np
from sklearn.base import BaseEstimator, OutlierMixin
from functools import reduce
from sklearn.base import BaseEstimator, OutlierMixin
from functools import reduce
from mtsa.features.mel import Array2Mfcc
from mtsa.models.GANF_components.ganfBaseModel import GANFBaseModel
from mtsa.utils import Wav2Array
from sklearn.pipeline import Pipeline
from mtsa.utils import Wav2Array

class GANF(nn.Module, BaseEstimator, OutlierMixin):

    def __init__(self, 
                 sampling_rate=None,
                 random_state = None,
                 isForWaveData = False,
                 batch_size = None,
                 learning_rate = None,
                 mono = True,
                 use_array2mfcc = False,
                 index_CUDA_device = '0'
                 ) -> None:
        super().__init__()
        self.sampling_rate = sampling_rate
        self.random_state = random_state
        self.isForWaveData =isForWaveData
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.mono = mono
        self.final_model = GANFBaseModel(index_CUDA_device=index_CUDA_device)
        self.use_array2mfcc = use_array2mfcc
        self.model = self._build_model()

    @property
    def name(self):
        return "GANF " + "+".join([f[0] for f in self.features])
        
    def fit(self, X, y=None, batch_size = None, epochs= None, max_iteraction= None, learning_rate = None, debug_dataframe = None):
        mono = (self.mono and not self.use_array2mfcc)
        if batch_size is None:
            batch_size = self.batch_size
        if learning_rate is None:
            learning_rate = self.learning_rate

        return self.model.fit(X, y, 
                              final_model__batch_size=batch_size,
                              final_model__epochs=epochs,
                              final_model__max_iteraction=max_iteraction,
                              final_model__learning_rate = learning_rate,
                              final_model__debug_dataframe = debug_dataframe,
                              final_model__isWaveData = self.isForWaveData,
                              final_model__mono = mono)

    def transform(self, X, y=None):
        l = list()
        l.append(X)
        l.extend(self.model.steps[:-1])
        Xt = reduce(lambda x, y: y[1].transform(x), l)
        return Xt
    
    def predict(self, X):
        return self.model.predict(X)

    def score_samples(self, X):
        return np.array(
            list(
                map(
                    self.model.score_samples, 
                    [[x] for x in X])
                )
            )
    
    def score(self, X, Y):
        X = self.score_samples
        fpr, tpr, thresholds = metrics.roc_curve(Y, X)
        auc = metrics.auc(fpr, tpr)
        return auc
    
    def get_adjacent_matrix(self):
        return self.final_model.get_adjacent_matrix()

    def _build_model(self):
        array2mfcc = Array2Mfcc(sampling_rate=self.sampling_rate)
        wav2array = Wav2Array(sampling_rate=self.sampling_rate, mono=self.mono)
        #mono = (self.mono and not self.use_array2mfcc)
        if self.use_array2mfcc and self.isForWaveData:
            model = Pipeline(
                steps=[
                    ("wav2array", wav2array),
                    ("array2mfcc", array2mfcc),
                    ("final_model", self.final_model),
                    ]
                )
        elif  self.isForWaveData:
            model = Pipeline(
                steps=[
                    ("wav2array", wav2array),
                    ("final_model", self.final_model),
                    ]
                )
        else:
            model = Pipeline(
                steps=[
                    ("final_model", self.final_model),
                    ]
                )   
            
        return model


from tensorflow.python.keras.layers import Dense, Layer

class Encoder(Layer):
    def __init__(self, latent_dim: int, activation: str, depth: int = 2, **kwargs,):
        super(Encoder, self).__init__(**kwargs)
        self.latent_dim = latent_dim
        self.activation = activation
        self.depth = depth
        self.kwargs = kwargs
        
    def build(self, input_shape):
        self.hidden = {
            'hidden_{}'.format(i): Dense(
                int(input_shape[-1] / (2**(i+1))), activation=self.activation,
            ) for i in range(self.depth)
        }
        self.latent = Dense(self.latent_dim, activation=self.activation)
        
    def call(self, inputs):
        x = self.hidden['hidden_0'](inputs)
        for i in range(1, self.depth):
            x = self.hidden['hidden_{}'.format(i)](x)
        return self.latent(x)
    
    def get_config(self):
        base_config = super(Encoder, self).get_config()
        config = {"latent_dim": self.latent_dim, "activation": self.activation,"depth": self.depth,}
        return dict(list(base_config.items()) + list(config.items()))


from tensorflow.python.keras.initializers import Constant
from tensorflow.python.keras.layers import Layer
from typing import List, Optional
from tensorflow.python.keras.constraints import NonNeg
import tensorflow as tf

class SinusoidalCoder(Layer):
    """ Fit m sinusoidal waves to an input t-matrix (matrix of m epochtimes) """
    def __init__(self, freq_init: Optional[List[float]] = None, max_freqs: int = 1, trainable_freq: bool = False, **kwargs):
        super(SinusoidalCoder, self).__init__(**kwargs)
        self.freq_init = freq_init
        if freq_init:
            self.max_freqs = len(freq_init)
        else:
            self.max_freqs = max_freqs
        self.trainable_freq = trainable_freq
        self.kwargs = kwargs
        
    def build(self, input_shape):
        self.amp = self.add_weight(shape=(input_shape[-1], self.max_freqs), initializer="zeros", trainable=True)
        if self.freq_init and not self.trainable_freq:
            self.freq = [self.add_weight(initializer=Constant(f), trainable=False) for f in self.freq_init]
        elif self.freq_init:
            self.freq = [self.add_weight(initializer=Constant(f), constraint=NonNeg(), trainable=True) for f in self.freq_init]
        else:
            self.freq = [
                self.add_weight(initializer="zeros", constraint=NonNeg(), trainable=True) for f in range(self.max_freqs)
            ]
        self.wb = self.add_weight(
            shape=(input_shape[-1], self.max_freqs), initializer="zeros", trainable=True
        )  # angular frequency (w) x phase shift
        self.disp = self.add_weight(shape=input_shape[-1], initializer="zeros", trainable=True)
    
    def call(self, inputs):
        th = tf.expand_dims(
            tf.expand_dims(self.freq, axis=0), axis=0
        ) * tf.expand_dims(inputs, axis=-1) + tf.expand_dims(self.wb, axis=0)
        return tf.reduce_sum(tf.expand_dims(self.amp, axis=0) * tf.sin(th), axis=-1) + self.disp
    
    def get_config(self):
        base_config = super(SinusoidalCoder, self).get_config()
        config = {"freq_init": self.freq_init, "max_freqs": self.max_freqs, "trainable_freq": self.trainable_freq}
        return dict(list(base_config.items()) + list(config.items()))


import numpy as np
from sklearn.base import BaseEstimator, OutlierMixin
from sklearn.pipeline import (
    Pipeline, 
    FeatureUnion
) 
from mtsa.features.stats import (
    MagnitudeMeanFeatureMfcc, 
    MagnitudeStdFeatureMfcc, 
    CorrelationFeatureMfcc,
    FEATURES,
    get_features
    )

from mtsa.features.mel import (
    Array2Mfcc 
)
from mtsa.utils import (
    Wav2Array,
)

from sklearn.mixture import GaussianMixture
from functools import reduce



FINAL_MODEL = GaussianMixture()

class MFCCMix(BaseEstimator, OutlierMixin):

    def __init__(self, 
                 final_model=FINAL_MODEL, 
                 features=FEATURES,
                 sampling_rate=None,
                 random_state = None,
                 ) -> None:
        super().__init__()
        self.sampling_rate = sampling_rate
        self.final_model = final_model
        self.random_state = random_state
        self.features = features
        self.model = self._build_model()

    @property
    def name(self):
        return "MFCCMix " + "+".join([f[0] for f in self.features])
        
    def fit(self, X, y=None):
        return self.model.fit(X, y)

    def transform(self, X, y=None):
        l = list()
        l.append(X)
        l.extend(self.model.steps[:-1])
        Xt = reduce(lambda x, y: y[1].transform(x), l)
        return Xt
    
    def predict(self, X):
        return self.model.predict(X)

    def score_samples(self, X):
        return self.model.score_samples(X=X)

    def _build_model(self):
        wav2array = Wav2Array(sampling_rate=self.sampling_rate)
        array2mfcc = Array2Mfcc(sampling_rate=self.sampling_rate)
        features = FeatureUnion(self.features)
        
        model = Pipeline(
            steps=[
                ("wav2array", wav2array),
                ("array2mfcc", array2mfcc),
                ("features", features),
                ("final_model", self.final_model),
                ]
            )
        
        return model

